name = 'pylibimport'
version = '0.1.2'
description = 'Python utility for importing packages with the same name, but different version.'
url = 'https://github.com/justengel/pylibimport'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'
