from .__meta__ import version as __version__

from .utils import make_import_name, get_name_version, is_python_package
from .get_versions import HttpListVersions
from .lib_import import VersionImporter
